@@ -24,19 +24,20 @@
 
 from collections import deque, defaultdict
 import functools
+import heapq
 import itertools
 import math
 import random
 import re
-import time
 import statistics
+import time
 
 from perlcompat import warn, die
 import numpy
 import pytess
 import tbdump
 
-VERSION = 1.7
+VERSION = 1.8
 
 CREATE_SUBP = {
     'random': 'create_random_graph',
@@ -59,23 +60,24 @@ CREATE_SUBP = {
     'configuration': 'create_configuration_graph',
     'regular': 'create_random_regular_graph',
     'li_maini': 'create_li_maini_graph',
+    'preset': 'create_preset_graph',
 }
 CREATE_TYPES = sorted(CREATE_SUBP.keys())
 
 IMPORT_SUBP = {
     'dot': 'import_dot',
+    'edges': 'import_edge_list',
+    'cell': 'import_cell',
 }
 IMPORT_FORMATS = sorted(IMPORT_SUBP.keys())
 
 EXPORT_SUBP = {
     'dot': 'export_dot',
+    'edges': 'export_edge_list',
     'cell': 'export_cell',
 }
 EXPORT_FORMATS = sorted(EXPORT_SUBP.keys())
 
-EPS = 1e-10
-INFINITY = 2 << 30
-
 MAX_RETRIES = 100
 
 class Graph:
@@ -180,10 +182,7 @@ class Graph:
 
     def has_vertex(self, v):
         """Check if vertex V exists in the graph."""
-        if self.V.get(v, None) is None:
-            return False
-        else:
-            return True
+        return v in self.V
 
     def add_vertex(self, v):
         """Attach vertex V to the graph if it does not exist."""
@@ -202,7 +201,7 @@ class Graph:
         if not ignore:
             self.expect_directed()
         # FIXME: should implement as a generator
-        if not self.EI.get(v, None):
+        if v not in self.EI:
             return []
         return self.EI[v].keys()
 
@@ -213,7 +212,7 @@ class Graph:
         if not ignore:
             self.expect_directed()
         # FIXME: Should implement as a generator.
-        if not self.EO.get(u, None):
+        if u not in self.EO:
             return []
         return self.EO[u].keys()
 
@@ -314,11 +313,9 @@ class Graph:
         """Check if the graph has edge (u, v)."""
         if self.undirected() and u > v:
             u, v = v, u
-        if not self.EO.get(u, None):
+        if u not in self.EO:
             return False
-        if not self.EO[u].get(v, None):
-            return False
-        return True
+        return v in self.EO[u]
 
     def get_multiedge_ids(self, u, v):
         """Return the edge identifiers (starting from zero) of edges between
@@ -345,14 +342,14 @@ class Graph:
             u, v = v, u
         count = self.get_edge_count(u, v)
         self.add_vertices(u, v)
-        if not self.EO.get(u, None):
+        if u not in self.EO:
             self.EO[u] = {}
-        if not self.EO[u].get(v, None):
+        if v not in self.EO[u]:
             self.EO[u][v] = {}
         self.EO[u][v][count] = {}  # default edge attributes
-        if not self.EI.get(v, None):
+        if v not in self.EI:
             self.EI[v] = {}
-        if not self.EI[v].get(u, None):
+        if u not in self.EI[v]:
             self.EI[v][u] = {}
         self.EI[v][u][count] = {}  # default edge attributes
         return count
@@ -491,6 +488,7 @@ class Graph:
         return self.get_edge_attribute_by_id(u, v, 0, 'weight')
 
     # algorithm ----------------------------------------------------------------
+    # https://stackoverflow.com/questions/22897209/dijkstras-algorithm-in-python
     def dijkstra(self, s):
         """Compute all shortest paths from source vertex S using Dijkstra's
         algorithm.  Return two dictionaries: DIST and PREV.  Dictionary DIST
@@ -501,32 +499,31 @@ class Graph:
         vertex in the shortest path from source vertex S to vertex 4.  You can
         obtain the shortest path to vertex V by traversing dictionary PREV
         from vertex V back to source node S."""
-        self.expect_undirected()
-        dist = {}
-        prev = {}
-        for v in self.vertices():
-            prev[v] = []
+        # Distance from the source to other vertices.
+        dist = {v: math.inf for v in self.vertices()}
         dist[s] = 0
+        prev = {v: [] for v in self.vertices()}
 
-        S = []
-        Q = self.vertices()
-        while Q:
-            Q = sorted(Q, key=lambda x: dist.get(x, INFINITY))
-            u = Q.pop(0)
-            if dist.get(u, None) is None:
-                break
-            S.append(u)
-            # NOTE: Use successors() for directeg graph.
-            for v in self.neighbors(u):
+        visited = {}
+        queue = []
+        heapq.heappush(queue, (dist[s], s))
+        while queue:
+            _, u = heapq.heappop(queue)
+            visited[u] = True
+            neighbors = self.neighbors(
+                u) if self.is_undirected() else self.successors(u)
+            for v in neighbors:
+                if v in visited:
+                    continue
                 # FIXME: Must reject multi-edged graph.
                 w = self.get_edge_weight_by_id(u, v, 0) or 1
-                if dist.get(v, None) is None or dist[v] > (
-                        dist.get(u, INFINITY) + w):
+                if dist[u] + w < dist[v]:  # Is vertex V closer from vertex U?
                     dist[v] = dist[u] + w
                     prev[v] = [u]
-                elif dist[v] == (dist.get(u, INFINITY) +
-                                 w):  # Handle equal paths.
+                elif dist[u] + w == dist[v]:  # Handle equal paths.
                     prev[v].append(u)
+                heapq.heappush(queue, (dist[u] + w, v))
+
         self.T[s] = dist
         self.P[s] = prev
         return dist, prev
@@ -559,7 +556,7 @@ class Graph:
         if t in self.T[s]:
             return self.T[s][t]
         else:
-            return INFINITY
+            return math.inf
 
     def dijkstra_all_pairs(self):
         """Compute all-pairs shortest paths using Dijkstra's algorithm."""
@@ -568,31 +565,33 @@ class Graph:
 
     def floyd_warshall(self):
         """Compute all-pairs shortest paths using Floyd-Warshall algorithm."""
-        # self.expect_directed()
         # Initialize weight matrix.
-        path = {}
+        dist = {}
         next_ = {}
         for v in self.vertices():
-            path[v] = {}
+            dist[v] = defaultdict(lambda: math.inf)
             next_[v] = {}
+
         for u, v in self.edges():
-            path[u][v] = self.get_edge_weight_by_id(u, v, 0) or 1
+            w = self.get_edge_weight_by_id(u, v, 0) or 1
+            dist[u][v] = w
+            if self.is_undirected():
+                dist[v][u] = w
 
         # Run Floyd-Warshall algorithm to find all-pairs shortest paths.
         for k in self.vertices():
             for u in self.vertices():
                 for v in self.vertices():
-                    if path[u].get(k, INFINITY) + path[k].get(
-                            v, INFINITY) < path[u].get(v, INFINITY):
-                        path[u][v] = path[u][k] + path[k][v]
+                    if dist[u][k] + dist[k][v] < dist[u][v]:
+                        dist[u][v] = dist[u][k] + dist[k][v]
                         next_[u][v] = k
-        self.T = path
+        self.T = dist
 
     def is_reachable(self, u, v):
         """Check if any path exists from vertex U to vertex V."""
-        if not self.T.get(u, None):
+        if u not in self.T:
             self.dijkstra(u)
-        return self.T[u].get(v, None)
+        return v in self.T[u]
 
     def explore(self, s):
         """Return the list of all vertices reachable from vertex S."""
@@ -1370,6 +1369,14 @@ class Graph:
 
         return self
 
+    def create_preset_graph(self, n=0):
+        self.import_edge_list("""\
+1 2
+2 3
+2 4
+3 4""".splitlines())
+        return self
+
     # import ----------------------------------------------------------------
     def import_graph(self, fmt, *args):
         name = IMPORT_SUBP.get(fmt, None)
@@ -1451,6 +1458,40 @@ class Graph:
                 self.add_vertex(v)
                 self.set_vertex_attributes(v, attrs)
 
+    def import_edge_list(self, lines):
+        """Creaete undirected graph from LINES in which every line describes
+        an edge (a pair of vertices) with its optional weight.  A prt of the
+        line after # are ignored.
+
+        Examples:
+        # unweighted K3 (complete graph with three vertices)
+        1 2
+        2 3
+        1 3
+
+        # weighted C4 (cubic)
+        1 2 1.5
+        2 3 1.1
+        3 4     # default to 1
+        4 1 2
+"""
+        for line in lines:
+            line = line.rstrip()
+            line = re.sub(r'#.*$', '', line)
+            u, v, *rest = line.split()
+            self.add_edge(u, v)
+            if rest:
+                w = float(rest[0])
+                self.set_edge_weight(u, v, w)
+
+    def import_cell(self, lines):
+        for line in lines:
+            line = line.rstrip()
+            m = re.search(r'define\s+e\d+\s+link\s+v(\d+)\s+v(\d+)', line)
+            if m:
+                u, v = map(int, m.groups())
+                self.add_edge(u, v)
+
     # ----------------------------------------------------------------
     def export_graph(self, fmt, *args):
         name = EXPORT_SUBP.get(fmt, None)
@@ -1471,8 +1512,6 @@ class Graph:
                 alist = []
                 for key, val in attrs.items():
                     alist.append(f'{key}="{val}"')
-                astr += ' [' + (', '.join(alist)) + ']'
-            astr += ';\n'
 
         for edge in sorted(self.unique_edges(), key=lambda e: e[0]):
             u = edge[0]
@@ -1492,18 +1531,27 @@ class Graph:
         astr += '}\n'
         return astr
 
+    def export_edge_list(self, *args):
+        astr = ''
+        for u, v in self.edges():
+            w = self.get_edge_weight(u, v)
+            if w:
+                astr += f'{u} {v} {w}\n'
+            else:
+                astr += f'{u} {v}\n'
+        return astr
+
     def export_cell(self, *args):
-        out = """\
-#define v_size 4 4
-#define v_color yellow
-#define e_width 1
-#define e_color blue
-"""
+        v_size = 12
+        v_color = 'white'
+        e_width = 1
+        e_color = 'blue'
+        out = ''
         for v in sorted(self.vertices()):
-            out += f'define v{v} ellipse v_size v_color\n'
+            out += f'define v{v} text {v} {v_size} {v_color}\n'
         n = 1
         for u, v in sorted(self.edges()):
-            out += f'define e{n} link v{u} v{v} e_width e_color\n'
+            out += f'define e{n} link v{u} v{v} {e_width} {e_color}\n'
             n += 1
 
         out += 'spring /^v/\n'
